"""TODO.

TODO.
"""

import numpy as np
from typing import NoReturn

__all__ = ["OLS"]


class OLS:
    """TODO."""

    def __init__(self) -> None:
        """TODO."""

        self.x_ = None
        self.y_ = None
        self.params_ = None
        self.residuals_ = None
        self.residuals_variance_ = None

    def fit(self, x: np.ndarray, y: np.ndarray) -> NoReturn:
        """TODO."""

        self.x_ = x
        self.y_ = y

        n = len(x)
        msg = f"x, y shapes must be ({n}, 1)"
        assert (x.shape == (n, 1)) and (y.shape == (n, 1)), msg

        x_matrix = self._build_x_matrix(x)
        x_matrix_t = np.transpose(x_matrix)
        self.params_ = np.matmul(np.matmul(np.linalg.inv(np.matmul(x_matrix_t,
                                                                   x_matrix)),
                                           x_matrix_t),
                                 y)
        self.residuals_ = self.forecast_error(x=x, y=y)
        residuals_variance_ = (np.matmul(np.transpose(self.residuals_),
                                         self.residuals_)
                               / (n - 2))
        self.residuals_variance_ = residuals_variance_[0, 0]

    @staticmethod
    def _build_x_matrix(x: np.ndarray) -> np.ndarray:
        """TODO."""

        n = len(x)
        msg = f"x shape must be ({n}, 1)"
        assert x.shape == (n, 1), msg
        x_ = np.concatenate((np.ones((n, 1)), x), axis=1)
        return x_

    def predict(self, x: np.ndarray) -> np.ndarray:
        """TODO."""

        y = np.matmul(self._build_x_matrix(x), self.params_)
        return y

    def forecast_error(self, x: np.ndarray, y: np.ndarray) -> np.ndarray:
        """TODO."""

        error = y - self.predict(x)
        return error

    def forecast_variance_matrix(self, x_new: np.ndarray) -> np.ndarray:
        """TODO."""

        n = len(x_new)
        x_matrix_new = self._build_x_matrix(x_new)
        x_matrix_new_t = np.transpose(x_matrix_new)
        x_matrix = self._build_x_matrix(self.x_)
        x_matrix_t = np.transpose(x_matrix)
        inv_x_t_x = np.linalg.inv(np.matmul(x_matrix_t, x_matrix))
        variance_matrix = (np.eye(n) * self.residuals_variance_
                           + (np.matmul(x_matrix_new,
                                        np.matmul(inv_x_t_x,
                                                  x_matrix_new_t))
                              * self.residuals_variance_))
        return variance_matrix



from event_studies.models.finance import MarketModel
import numpy as np
from typing import List

__all__ = ["calculate_aggregations", "j1", "j2", "rank_test",
           "generalized_sign_test"]


def calculate_aggregations(market_model: MarketModel,
                           event_sample: List[dict]) -> List[dict]:
    """TODO."""

    aggregations = []
    for event in event_sample:
        market_model.fit(firm_estimation_data=event["estimation_data"],
                         price_col=event["firm"],
                         date_col=event["date_col"])
        event_agg = market_model.abnormal_return_aggregations(event["start"],
                                                              event["end"],
                                                              firm_event_data=event["event_data"],
                                                              price_col=event["firm"],
                                                              date_col=event["date_col"])
        aggregations.append(event_agg)

    return aggregations


def j1(aggregations) -> float:
    """TODO."""

    cars = []
    var_cars = []
    for aggregation in aggregations:
        cars.append(aggregation["parametric"]["car"])
        var_cars.append(aggregation["parametric"]["var_car"])

    average_car = np.array(cars).mean()
    var_average_car = np.array(var_cars).mean() / len(aggregations)
    j1_value = average_car / np.sqrt(var_average_car)

    return j1_value


def j2(aggregations) -> float:
    """TODO."""

    scars = []
    n = len(aggregations)
    dfs = []
    for aggregation in aggregations:
        scars.append(aggregation["parametric"]["scar"])
        dfs.append(aggregation["parametric"]["scar_t_student_df"])

    msg = "different estimation window size for events not allowed!"
    assert all(x == dfs[0] for x in dfs), msg
    df = dfs[0]

    average_scar = np.array(scars).mean()
    normalization_constant = np.sqrt(n * (1 - 2/df))
    j2_value = normalization_constant * average_scar

    return j2_value


def rank_test(aggregations) -> float:
    """TODO."""

    event_ranks = []
    ws = []
    total_ranks = []
    for aggregation in aggregations:
        ws.append(aggregation["event_window_size"])
        total_ranks.append(len(aggregation["ar_total_midranks"]))
        event_ranks.append(aggregation["ar_total_midranks"][-ws[-1]:, 0])

    msg = "different event window size for events not allowed!"
    assert all(x == ws[0] for x in ws), msg

    msg = "different event+estimation window size for events not allowed!"
    assert all(x == total_ranks[0] for x in total_ranks), msg

    event_ranks = np.array(event_ranks)
    average_rank_on_event = event_ranks.mean()
    expected_rank = (total_ranks[0] + 1) / 2
    rank_variance = ((event_ranks.mean(axis=0) - expected_rank) ** 2).mean()
    rank_statistic = (np.sqrt(ws[0]) * (average_rank_on_event - expected_rank)
                      / np.sqrt(rank_variance))

    return rank_statistic


def generalized_sign_test(aggregations) -> float:
    """TODO."""

    proportions = []
    w = 0
    for aggregation in aggregations:
        proportions.append(aggregation["estimation_window_ar_sign"])
        if aggregation["parametric"]["car"] > 0.0:
            w += 1

    p = np.mean(proportions)
    n = len(proportions)
    generalized_sign_statistic = (w - n * p) / np.sqrt(n * p * (1-p))

    return generalized_sign_statistic



"""TODO.

TODO.
"""

import numpy as np
import pandas as pd
import pandas.api.types as ptypes
from event_studies.models.linear_models import OLS
from typing import NoReturn, List, Optional

__all__ = ["MarketModel"]


class MarketModel:
    """TODO."""

    _date_col = "Date"

    def __init__(self, market_data: pd.DataFrame, market_index_col: str,
                 date_col: Optional[str] = None) -> None:
        """TODO."""

        if date_col is None:
            date_col = self._date_col

        self.market_index_col = market_index_col
        self._market_data = self._read_time_series(time_series=market_data,
                                                   data_col=market_index_col,
                                                   date_col=date_col)
        self.ols_ = None
        self.data_ = None
        self.firm_col_ = None

    @classmethod
    def _read_time_series(cls, time_series: pd.DataFrame, data_col: str,
                          date_col: str,
                          forbidden_names: List[str] = None) -> pd.DataFrame:
        """TODO."""

        if forbidden_names is None:
            forbidden_names = []

        forbidden_names.append(cls._date_col)
        msg = f"{data_col} is an invalid name for a data column"
        assert data_col not in forbidden_names, msg

        msg = f"{date_col} is not of datetime type"
        assert ptypes.is_datetime64_any_dtype(time_series[date_col]), msg
        msg = f"{data_col} is not numeric"
        assert ptypes.is_numeric_dtype(time_series[data_col]), msg

        time_series_ = time_series[[date_col, data_col]].copy()
        time_series_.rename(columns={date_col: cls._date_col}, inplace=True)

        return time_series_

    def _load_data(self, firm_estimation_data: pd.DataFrame, price_col: str,
                   date_col: str) -> NoReturn:
        """TODO."""

        self.data_ = self._data_plus_market(firm_data=firm_estimation_data,
                                            data_col=price_col,
                                            date_col=date_col)
        self.firm_col_ = price_col

    def fit(self, firm_estimation_data: pd.DataFrame, price_col: str,
            date_col: str = "Date") -> NoReturn:
        """TODO."""

        self._load_data(firm_estimation_data=firm_estimation_data,
                        price_col=price_col, date_col=date_col)
        self.ols_ = OLS()
        x = self.calculate_returns(prices=self.data_[self.market_index_col])
        y = self.calculate_returns(prices=self.data_[price_col])
        self.ols_.fit(x=x, y=y)

    def _data_plus_market(self, firm_data: pd.DataFrame, data_col: str,
                          date_col: str) -> pd.DataFrame:
        """TODO."""

        forbidden_names = [self.market_index_col]
        firm_data_ = self._read_time_series(time_series=firm_data,
                                            data_col=data_col,
                                            date_col=date_col,
                                            forbidden_names=forbidden_names)
        data_ = firm_data_.merge(self._market_data, on=self._date_col,
                                 how="left")\
                          .sort_values(by=self._date_col, ascending=True)\
                          .fillna(method="ffill", axis="columns")

        msg = "Data not available for the market portfolio on the given dates"
        assert data_[self.market_index_col].notna().all(), msg

        data_.set_index(self._date_col, inplace=True)
        return data_

    @staticmethod
    def calculate_returns(prices: pd.Series) -> np.ndarray:
        """TODO."""

        returns = np.log(((prices / prices.shift(1)).iloc[1:]).astype(float))
        n = len(returns)
        returns = returns.to_numpy().reshape(n, 1)

        return returns

    def _predict(self, event_data: pd.DataFrame) -> np.ndarray:
        """TODO."""

        market_index = event_data[self.market_index_col]
        event_market_returns = self.calculate_returns(prices=market_index)
        return self.ols_.predict(x=event_market_returns)

    def abnormal_returns(self, firm_event_data: pd.DataFrame, price_col: str,
                         date_col: str) -> np.ndarray:
        """TODO."""

        event_data = self._data_plus_market(firm_data=firm_event_data,
                                            data_col=price_col,
                                            date_col=date_col)
        market_index = event_data[self.market_index_col]
        market_returns = self.calculate_returns(market_index)
        firm_returns = self.calculate_returns(event_data[price_col])
        abnormal_returns = self.ols_.forecast_error(market_returns, firm_returns)
        return abnormal_returns

    def abnormal_returns_variance(self, firm_event_data: pd.DataFrame,
                                  price_col: str,
                                  date_col: str) -> np.ndarray:
        """TODO."""

        event_data = self._data_plus_market(firm_data=firm_event_data,
                                            data_col=price_col,
                                            date_col=date_col)
        market_index = event_data[self.market_index_col]
        market_returns = self.calculate_returns(market_index)
        variance_matrix = self.ols_.forecast_variance_matrix(market_returns)
        return variance_matrix

    def abnormal_return_aggregations(self, window_start: int, window_end: int,
                                     *args, **kwargs) -> dict:
        """TODO."""

        abnormal_returns = self.abnormal_returns(*args, **kwargs)
        abnormal_returns_variance = self.abnormal_returns_variance(*args,
                                                                   **kwargs)
        aggregations = {}
        l2 = len(abnormal_returns)
        event_window_size = window_end - window_start + 1
        gamma = np.concatenate((np.zeros((window_start, 1)),
                                np.ones((event_window_size, 1)),
                                np.zeros((l2 - window_end - 1, 1))), axis=0)

        car = np.matmul(np.transpose(gamma), abnormal_returns)[0, 0]
        var_car = np.matmul(np.transpose(gamma),
                            np.matmul(abnormal_returns_variance, gamma))[0, 0]
        scar = car / np.sqrt(var_car)
        aggregations["event_window_size"] = event_window_size
        aggregations["parametric"] = {"car": car, "scar": scar,
                                      "var_car": var_car,
                                      "scar_t_student_df": (len(self.ols_.x_)
                                                            - 2)}
        excess_returns = np.concatenate((self.ols_.residuals_,
                                         abnormal_returns))
        # for the rank test
        aggregations["ar_total_midranks"] = self._midranks(excess_returns)
        # for the sign test
        name = "estimation_window_ar_sign"
        aggregations[name] = np.where(self.ols_.residuals_ > 0,
                                      1.0, 0.0).mean()

        return aggregations

    @staticmethod
    def _midranks(values: np.ndarray) -> np.ndarray:
        """TODO."""

        n = len(values)
        assert values.shape == (n, 1), "Shape not allowed!"

        temp = values.argsort(axis=0)
        ranks = np.zeros(temp.shape, dtype=float)

        mid_ranks = []
        k_smaller = 0
        l_ties = 1
        last_number = values.min()

        for iter_count, idx in enumerate(temp[1:, 0]):
            num = values[idx, 0]
            if num > last_number:
                mid_ranks += list(np.repeat(k_smaller + (1 + l_ties) / 2,
                                            l_ties))
                # flush
                l_ties = 1
                k_smaller = iter_count + 1
                last_number = num
            else:
                l_ties += 1

        mid_ranks += list(np.repeat(k_smaller + (1 + l_ties) / 2, l_ties))
        ranks[temp] = np.array(mid_ranks).reshape(len(mid_ranks), 1, 1)
        return ranks
